@@ -63,6 +63,22 @@ using namespace KAStats::Terms;
 
 K_PLUGIN_FACTORY( SidebarModeFactory, registerPlugin<SidebarMode>(); )
 
+FocusHackWidget::FocusHackWidget(QWidget *parent)
+    : QWidget(parent)
+{}
+FocusHackWidget::~FocusHackWidget()
+{}
+
+void FocusHackWidget::focusNext()
+{
+    focusNextChild();
+}
+
+void FocusHackWidget::focusPrevious()
+{
+    focusNextPrevChild(false);
+}
+
 class SubcategoryModel : public QStandardItemModel
 {
 public:
@@ -221,7 +237,7 @@ public:
     KPackage::Package package;
     SubcategoryModel * subCategoryModel;
     MostUsedModel * mostUsedModel;
-    QWidget * mainWidget;
+    FocusHackWidget * mainWidget;
     QQuickWidget * placeHolderWidget;
     QHBoxLayout * mainLayout;
     KDeclarative::KDeclarative kdeclarative;
@@ -247,6 +263,8 @@ SidebarMode::SidebarMode( QObject *parent, const QVariantList& )
     d->aboutIcon->addAuthor( i18n( "Ben Cooksley" ), i18n( "Author" ), "bcooksley@kde.org" );
     d->aboutIcon->addAuthor( i18n( "Mathias Soeken" ), i18n( "Developer" ), "msoeken@informatik.uni-bremen.de" );
     d->aboutIcon->setProgramIconName( "view-sidetree" );
+
+    qmlRegisterType<QAction>();
 }
 
 SidebarMode::~SidebarMode()
@@ -320,7 +338,7 @@ void SidebarMode::initEvent()
     d->mostUsedModel = new MostUsedModel( this );
 
     d->subCategoryModel = new SubcategoryModel( d->searchModel, this );
-    d->mainWidget = new QWidget();
+    d->mainWidget = new FocusHackWidget();
     d->mainWidget->installEventFilter(this);
     d->mainLayout = new QHBoxLayout(d->mainWidget);
     d->mainLayout->setContentsMargins(0, 0, 0, 0);
@@ -330,16 +348,22 @@ void SidebarMode::initEvent()
     moduleView()->setFaceType(KPageView::Plain);
 }
 
-void SidebarMode::triggerGlobalAction(const QString &name)
+QAction *SidebarMode::action(const QString &name) const
 {
     if (!d->collection) {
-        return;
+        return nullptr;
     }
 
-    QAction *action = d->collection->action(name);
-    if (action) {
-        action->trigger();
+    return d->collection->action(name);
+}
+
+QString SidebarMode::actionIconName(const QString &name) const
+{
+    if (QAction *a = action(name)) {
+        return a->icon().name();
     }
+
+    return QString();
 }
 
 void SidebarMode::requestToolTip(int index, const QRectF &rect)
@@ -439,7 +463,9 @@ void SidebarMode::initWidget()
     }
 
     d->quickWidget = new QQuickWidget(d->mainWidget);
+    d->quickWidget->quickWindow()->setTitle(i18n("Sidebar"));
     d->quickWidget->setResizeMode(QQuickWidget::SizeRootObjectToView);
+
     d->quickWidget->engine()->rootContext()->setContextProperty("systemsettings", this);
     d->package = KPackage::PackageLoader::self()->loadPackage("KPackage/GenericQML");
     d->package.setPath(QStringLiteral("org.kde.systemsettings.sidebar"));
@@ -464,16 +490,22 @@ void SidebarMode::initWidget()
                     d->quickWidget->setFixedWidth(240);
                 }
             });
+    connect(d->quickWidget->rootObject(), SIGNAL(focusNextRequest()), d->mainWidget, SLOT(focusNext()));
+    connect(d->quickWidget->rootObject(), SIGNAL(focusPreviousRequest()), d->mainWidget, SLOT(focusPrevious()));
 
     d->quickWidget->installEventFilter(this);
 
     d->toolTipManager = new ToolTipManager(d->searchModel, d->quickWidget);
 
     d->placeHolderWidget = new QQuickWidget(d->mainWidget);
+    d->placeHolderWidget->quickWindow()->setTitle(i18n("Most Used"));
     d->placeHolderWidget->setResizeMode(QQuickWidget::SizeRootObjectToView);
     d->placeHolderWidget->engine()->rootContext()->setContextObject(new KLocalizedContext(d->placeHolderWidget));
     d->placeHolderWidget->engine()->rootContext()->setContextProperty("systemsettings", this);
     d->placeHolderWidget->setSource(QUrl::fromLocalFile(d->package.filePath("ui", "introPage.qml")));
+    connect(d->placeHolderWidget->rootObject(), SIGNAL(focusNextRequest()), d->mainWidget, SLOT(focusNext()));
+    connect(d->placeHolderWidget->rootObject(), SIGNAL(focusPreviousRequest()), d->mainWidget, SLOT(focusPrevious()));
+    d->placeHolderWidget->installEventFilter(this);
 
     d->mainLayout->addWidget( d->quickWidget );
     d->moduleView->hide();
@@ -486,8 +518,28 @@ void SidebarMode::initWidget()
 
 bool SidebarMode::eventFilter(QObject* watched, QEvent* event)
 {
-    //TODO: patch Qt
-    if (watched == d->quickWidget && event->type() == QEvent::Leave) {
+    //FIXME: those are all workarounds around the QQuickWidget brokeness
+    if ((watched == d->quickWidget || watched == d->placeHolderWidget)
+         && event->type() == QEvent::KeyPress) {
+        //allow tab navigation inside the qquickwidget
+        QKeyEvent *ke = static_cast<QKeyEvent *>(event);
+        QQuickWidget *qqw = static_cast<QQuickWidget *>(watched);
+        if (ke->key() == Qt::Key_Tab || ke->key() == Qt::Key_Backtab) {
+            QCoreApplication::sendEvent(qqw->quickWindow(), event);
+            return true;
+        }
+    } else if ((watched == d->quickWidget || watched == d->placeHolderWidget)
+                && event->type() == QEvent::FocusIn) {
+        QFocusEvent *fe = static_cast<QFocusEvent *>(event);
+        QQuickWidget *qqw = static_cast<QQuickWidget *>(watched);
+        if (qqw && qqw->rootObject()) {
+            if (fe->reason() == Qt::TabFocusReason) {
+                QMetaObject::invokeMethod(qqw->rootObject(), "focusFirstChild");
+            } else if (fe->reason() == Qt::BacktabFocusReason) {
+                QMetaObject::invokeMethod(qqw->rootObject(), "focusLastChild");
+            }
+        }
+    } else if (watched == d->quickWidget && event->type() == QEvent::Leave) {
         QCoreApplication::sendEvent(d->quickWidget->quickWindow(), event);
     } else if (watched == d->mainWidget && event->type() == QEvent::Resize) {
         emit widthChanged();
